# -*- coding: utf-8 -*-
"""Export predictions to PineAPPL."""
import json

import numpy as np

import yadism


def dump_pineappl_to_file(output, filename, obsname):
    """Write output on a PineAPPL grid file.

    Parameters
    ----------
    filename : str
        output file name
    obsname : str
        observable to be dumped

    """
    # pylint: disable=no-member, too-many-locals
    if len(output[obsname]) <= 0:
        raise ValueError(f"no ESF {obsname}!")
    import pineappl  # pylint: disable=import-outside-toplevel,import-error

    interpolation_xgrid = output["interpolation_xgrid"]
    # interpolation_is_log = self["interpolation_is_log"]
    interpolation_polynomial_degree = output["interpolation_polynomial_degree"]
    lepton_pid = output["projectilePID"]

    # init pineappl objects
    lumi_entries = [
        pineappl.lumi.LumiEntry([(pid, lepton_pid, 1.0)]) for pid in output["pids"]
    ]
    first_esf_result = output[obsname][0]
    orders = [pineappl.grid.Order(*o) for o in first_esf_result.orders]
    bins = len(output[obsname])
    bin_limits = list(map(float, range(0, bins + 1)))
    # subgrid params
    params = pineappl.subgrid.SubgridParams()
    params.set_reweight(False)
    params.set_x_bins(len(interpolation_xgrid))
    params.set_x_max(interpolation_xgrid[-1])
    params.set_x_min(interpolation_xgrid[0])
    params.set_x_order(interpolation_polynomial_degree)

    grid = pineappl.grid.Grid.create(lumi_entries, orders, bin_limits, params)
    limits = []

    # add each ESF as a bin
    for bin_, obs in enumerate(output[obsname]):
        x = obs.x
        Q2 = obs.Q2

        limits.append((Q2, Q2))
        limits.append((x, x))

        # add all orders
        for o, (v, _e) in obs.orders.items():
            order_index = list(first_esf_result.orders.keys()).index(o)
            prefactor = (
                ((1.0 / (4.0 * np.pi)) ** o[0]) * ((-1.0) ** o[2]) * ((-1.0) ** o[3])
            )
            # add for each pid/lumi
            for pid_index, pid_values in enumerate(v):
                pid_values = prefactor * pid_values
                # grid is empty? skip
                if not any(np.array(pid_values) != 0):
                    continue
                subgrid = pineappl.import_only_subgrid.ImportOnlySubgridV1(
                    pid_values[np.newaxis, :, np.newaxis],
                    [Q2],
                    interpolation_xgrid,
                    [1.0],
                )
                grid.set_subgrid(order_index, bin_, pid_index, subgrid)
    # set the correct observables
    normalizations = [1.0] * bins
    remapper = pineappl.bin.BinRemapper(normalizations, limits)
    grid.set_remapper(remapper)

    # set the initial state PDF ids for the grid
    grid.set_key_value("initial_state_1", "2212")
    grid.set_key_value("initial_state_2", str(lepton_pid))
    grid.set_key_value("theory", json.dumps(output.theory))
    grid.set_key_value("runcard", json.dumps(output.observables))
    grid.set_key_value("yadism_version", yadism.__version__)
    grid.set_key_value("lumi_id_types", "pdg_mc_ids")
    # set bin information
    grid.set_key_value("x1_label", "Q2")
    grid.set_key_value("x1_label_tex", "$Q^2$")
    grid.set_key_value("x1_unit", "GeV^2")
    grid.set_key_value("x2_label", "x")
    grid.set_key_value("x2_label_tex", "$x$")
    grid.set_key_value("x2_unit", "")
    grid.set_key_value("y_label", obsname)

    # dump file
    grid.optimize()
    grid.write(filename)



# -*- coding: utf-8 -*-
"""
Note that at |N3LO| the source files in fortran follow :cite:`vogt-f3cc` notation
where the odd-N moments are called minus even if they correspond to :math:`\nu + \bar{\nu}`.
This convention is changed in :cite:`Davies:2016ruz` where the |N3LO| CC results are presented
for the first time. Referred equations are not always in agreement with the code conventions.
"""
from .. import partonic_channel as pc
from ..partonic_channel import RSL
from . import f3_nc, n3lo, nnlo


class NonSingletOdd(f3_nc.NonSinglet):
    def NNLO(self):
        """
        |ref| implements :eqref:`3.6`, :cite:`vogt-f3cc`
        or :eqref:`2.8`, :cite:`Davies:2016ruz`.
        """

        return RSL(
            nnlo.xc3ns2p.c3nm2a, nnlo.xc3ns2p.c3ns2b, nnlo.xc3ns2p.c3nm2c, [self.nf]
        )

    def N3LO(self):
        """
        |ref| implements :eqref:`3.7`, :cite:`vogt-f3cc` or :eqref:`2.11`, :cite:`Davies:2016ruz`.
        """

        return RSL(
            n3lo.xc3ns3p.c3nm3a,
            n3lo.xc3ns3p.c3ns3b,
            n3lo.xc3ns3p.c3nm3c,
            [self.nf, False],
        )


class NonSingletEven(f3_nc.NonSinglet):
    def N3LO(self):
        """
        |ref| implements the sum between :eqref:`2.11` and :eqref:`3.8`, :cite:`Davies:2016ruz`.
        """

        return RSL(
            n3lo.xc3ns3p.c3np3a,
            n3lo.xc3ns3p.c3ns3b,
            n3lo.xc3ns3p.c3np3c,
            [self.nf, False],
        )


class Gluon(pc.EmptyPartonicChannel):
    pass


class Singlet(pc.EmptyPartonicChannel):
    pass



# -*- coding: utf-8 -*-
"""
See :mod:`f3_cc` docstring for the name conventions.
"""
from .. import partonic_channel as pc
from ..partonic_channel import RSL
from . import f2_nc, n3lo, nlo, nnlo


class NonSinglet(f2_nc.NonSinglet):
    @staticmethod
    def NLO():
        """
        |ref| implements :eqref:`155`, :cite:`moch-f3nc`.
        """

        return RSL.from_distr_coeffs(
            nlo.f3.ns_reg, (nlo.f2.ns_delta, nlo.f2.ns_omx, nlo.f2.ns_logomx)
        )

    def NNLO(self):
        """
        |ref| implements the sum between :eqref:`2.8` and :eqref:`3.5`, :cite:`Davies:2016ruz`
        or :eqref:`208`, :cite:`moch-f3nc`.
        """

        return RSL(
            nnlo.xc3ns2p.c3np2a, nnlo.xc3ns2p.c3ns2b, nnlo.xc3ns2p.c3np2c, [self.nf]
        )

    def N3LO(self):
        """
        |ref| implements the sum between :eqref:`2.11` and :eqref:`3.8`, :cite:`Davies:2016ruz`.
        """

        return RSL(
            n3lo.xc3ns3p.c3np3a,
            n3lo.xc3ns3p.c3ns3b,
            n3lo.xc3ns3p.c3np3c,
            [self.nf, True],
        )


class Gluon(pc.EmptyPartonicChannel):
    pass


class Singlet(pc.EmptyPartonicChannel):
    pass
